/* --------------------------------------------------------------
    Integral arithmetical calculator.

    http://www.imach.uran.ru/cbignum

    Copyright 2002-2007 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifdef  _MSC_VER
#pragma warning(disable:4100)
#endif/*_MSC_VER*/
#include <assert.h>
#include <ctype.h>
#include <fstream.h>
#include <iomanip.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "Gettimer.h"
#include "Random.h"
#include <NTL/ZZ.h>

static  int GenerateRandom (ZZ& a, size_t radix, size_t maxexp);
static  int GenerateRandom (long a, size_t radix, size_t maxexp);
static void ShowStat (ZZ& a, ZZ& na);

long    SetTimer (long &timer)
{
  long temp = GetTimer(0);
  long time = temp - timer;
  timer = temp;

  return (time >= 0? time: 0L);
}

const char szID[] =
    "Big integer calculator. Version 1.2a NTL beta public of "__DATE__"\n"
    "Copyright 2002-2007 by Dr. Raul N.Shakirov, IMach of RAS(UB)\n"
    ;
const char szHelp[] =
    "Call:  Arifexp [Options]\n"
    "Options:\n"
    "-idle   idle priority\n"
    "-high   idle priority\n"
    "-hex    hexadecimal input/output\n"
    "-hexi   hexadecimal input\n"
    "-hexo   hexadecimal output\n"
    "-div0   allow division by 0\n"
    "-size   show high word and size on input numbers\n"
    "-v      write expression before the result\n"
    "-rep n  repeat operation n times\n"
    "-check  check results of operations, if possible\n"
    "-rand   generate random numbers instead of computing:\n"
    "        positive number of length n instead of value n\n"
    "        negative number of length n instead of value -n\n"
    "        except for degree of shift and power, unless option -rand all\n"
    "-exp n  append up to n trailing zeroes to random numbers\n"
    "-help   show this help and exit\n"
    "\n"
    "Site:   http://www.imach.uran.ru/cbignum\n"
    ;

int main (int argc, char** argv)
{
  int radix    = 10;
  int nSize    = 0;
  int nVerbose = 0;
  int nRepeat  = 1;
  int nCheck   = 0;
  int maxexp   = 0;

  int retcode = 0;
  long total_r = 0;
  long total_c = 0;
  long total_w = 0;

  cin >> dec;
  cout << dec;
  cerr << szID << endl;

  // Look for arguments.

  char *pszPriority = "";
  char *pszGenerate = "";
  char *pszAll      = "";

  for (; argc > 1; --argc, ++argv)
  {
    if (strcmp (argv [1], "-idle") == 0)
    {
      if (SetIdlePriority()) pszPriority = "Idle priority\n\n";
      continue;
    }
    if (strcmp (argv [1], "-high") == 0)
    {
      if (SetHighPriority()) pszPriority = "High priority\n\n";
      continue;
    }
    if (strcmp (argv [1], "-hex") == 0)
    {
      cin >> hex;
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-hexi") == 0)
    {
      cin >> hex;
      continue;
    }
    if (strcmp (argv [1], "-hexo") == 0)
    {
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-div0") == 0)
    {
//    ZZ::maskdiv0 (1);
      continue;
    }
    if (strcmp (argv [1], "-size") == 0)
    {
      nSize = 1;
      continue;
    }
    if (strcmp (argv [1], "-v") == 0)
    {
      nVerbose = 1;
      continue;
    }
    if (strcmp (argv [1], "-rep") == 0 && argc > 2)
    {
      nRepeat = atoi (argv [2]);
      --argc, ++argv;
      continue;
    }
    if (strcmp (argv [1], "-check") == 0)
    {
      nCheck = 1;
      continue;
    }
    if (strcmp (argv [1], "-rand") == 0)
    {
      RRand3();
      pszGenerate = "Generating random numbers\n\n";
      if (argc > 2 && strcmp (argv [2], "all") == 0)
      {
        pszAll = "all";
        --argc, ++argv;
      }
      continue;
    }
    if (strcmp (argv [1], "-exp") == 0 && argc > 2)
    {
      maxexp = atoi (argv [2]);
      --argc, ++argv;
      continue;
    }
    if (strcmp (argv [1],  "-help") == 0 ||
        strcmp (argv [1],  "-h")    == 0 ||
        strcmp (argv [1],  "-?")    == 0)
    {
      cout << szHelp;
      return 0;
    }

    // Unrecognized option.

    cerr << "Unrecognized option " << argv [1] << endl;
    return 1;
  }

  cerr << pszPriority;
  cerr << pszGenerate;

  ZZ a;         // Operand 1
  ZZ b;         // Operand 2
  ZZ c;         // Operand 3
  ZZ null;      // Null number
  ZZ x;         // Result 1
  ZZ y;         // Result 2
  ZZ p;         // Result to compare
  ZZ q;         // Result to compare
  ZZ cmp;       // Operand to compare
  ZZ na;        // Number of bits in a
  ZZ nb;        // Number of bits in b
  ZZ nc;        // Number of bits in c
  ZZ nt;        // Estimated number of CPU cycles to calculate

  while (cin)
  {
    cerr << "Please, type expression "
         << "a=, a+-*/%\\|&^b=, a+-b*c=, a\\b%c=, a<<>>b= or 2Va=\n"
         << "and press Enter or press Enter to exit\n";

    long timer = GetTimer (0);

    char ch = 0; cin.get (ch);
    if (cin.eof() || ch == '\n') break;
    cin.putback (ch);

    a = 0;              // Operand 1
    b = 0;              // Operand 2
    c = 0;              // Operand 3
    null = 0;           // Null number
    int pp = ch;        // First symbol
    int op = 0, md = 0; // Operators
    unsigned error = 0; // Error mask
    unsigned alt = 1;   // Use alternative operation in check mode
    unsigned rev = 1;   // Use reverse operation in check mode

    cin >> a;
    if (!cin) { error &= 0x0010; cin.clear(); cin >> null; }
    cin >> ch;

    switch (ch)
    {
      case '<': case '>': op = ch; ch = 0; cin >> ch;
                          if (op != ch) op = ch = '?';
    }

    if (!cin.eof() && ch != '=')
    {
      op = ch; ch = 0; cin >> b;
      if (!cin) { error &= 0x0020; cin.clear(); cin >> null; }
      cin >> ch;
    }

    if (!cin.eof() && ch != '=')
    {
      md = ch; ch = 0; cin >> c;
      if (!cin) { error &= 0x0040; cin.clear(); cin >> null; }
      cin >> ch;
    }

    na = NumBits(a);
    nb = NumBits(b);
    nc = NumBits(c);
    nt = 0;
    long lb = trunc_long(b,CHAR_BIT * sizeof (long));

    if (op != 0)
    {
      if (md == 0)
      switch (op)
      {
        case '+':   if (pp == op && a == 0) { op = +1; a = b; break; }
                    nt = na; if (nt < nb) nt = nb; (++nt) /= 15;
                    break;
        case '-':   if (pp == op && a == 0) { op = -1; a = b; break; }
                    nt = na; if (nt < nb) nt = nb; (++nt) /= 15;
                    break;
        case '*':   if (a == 0 || b == 0) rev = 0;
        case '/':
        case '%':   nt = na * nb / 30;
                    alt = 0; break;
        case '\\':  alt = 0;
                    switch (lb) {
                      case 8: case 4: case 2: case 1: break;
                      default: rev = 0;
                    }        break;
        case '<':            break;
        case '>':   rev = 0; break;
        case '|':   alt = 0; rev = 0; break;
        case '&':   alt = 0; rev = 0; break;
        case '^':   alt = 0;          break;
        case 'V':
        case 'v':   nt = nb * nb / 90;
                    alt = 0; if (a == 2) break;
        default:    error = 0x0001;
      }
      else if (md == '*')
      switch (op)
      {
        case '+':
        case '-':   if (b == 0 || c == 0) rev = 0;
                    alt = 0; break;
        default:    error = 0x0001;
      }
      else if (md == '%')
      switch (op)
      {
        case '\\':  null = na; null *= b; null /= 8;
                    if (null > 0x00FFFFFFUL / 2)
                        alt = 0;
                    rev = 0; break;
        default:    error = 0x0001;
      }
      else error = 0x0002;
    }

    while (cin && !cin.eof() && ch != '=')
    {
      error |= 0x4000; cin >> ch;               // Skip until =
    }

    while (cin && !cin.eof() && (cin.get (ch), ch != '\n'))
    {
      error |= (isspace (ch) - 1) & 0x8000;     // Skip rest of line
    }

    if (error & 0x0001)
      cout << "Error: 1st operator is either missed or unknown" << endl;

    if (error & 0x0002)
      cout << "Error: 2nd operator is either missed or unknown" << endl;

    if (error & 0x0010)
      cout << "Error: 1st operand too long" << endl;

    if (error & 0x0020)
      cout << "Error: 2nd operand too long" << endl;

    if (error & 0x0040)
      cout << "Error: 3rd operand too long" << endl;

    if (error & 0x4000)
      cout << "Error: excessive operators or operands before sign =" << endl;

    if (error & 0x8000)
      cout << "Error: misplaced operators or operands after sign =" << endl;

    if (error)
    {
      retcode = 1;
      cout << endl;
      continue;
    }

    total_r += SetTimer (timer);

    // Generating random numbers.

    if (*pszGenerate != 0)
    {
      switch (op)
      {
        case  +1:   op = '+'; cout << (char) op;
                    break;
        case  -1:   op = '-'; cout << (char) op;
                    break;
        case 'V':
        case 'v':   cout << a << endl;
                    break;
        default:    retcode |= GenerateRandom (a, radix, maxexp);
                    cout << endl;
      }

      if (op != 0)
      {
        cout << (char) op;
        switch (op)
        {
          case '<':
          case '>': cout << (char) op;
          case '\\':if (*pszAll == 0)
                    {
                       cout << endl << b;
                       break;
                    }
          default:  cout << endl;
                    retcode |= GenerateRandom (b, radix, maxexp);
        }
        if (md != 0)
        {
          cout << endl << (char) md << endl;
          retcode |= GenerateRandom (c, radix, maxexp);
        }
        cout << endl;
      }
      cout << '=' << endl;
      continue;
    }

    // Reporting size of numbers.

    if (nSize)
    {
      switch (op)
      {
        case  +1:   break;
        case  -1:   break;
        case 'V':   break;
        case 'v':   break;
        default:    ShowStat (a, na);
      }
      if (op != 0)  ShowStat (b, nb);
      if (md != 0)  ShowStat (c, nc);
    }

    // Writing operands.

    if (nVerbose != 0)
    {
      switch (op)
      {
        case  +1:   op = '+'; cout << (char) op;
                    break;
        case  -1:   op = '-'; cout << (char) op;
                    break;
        default:    cout << a << endl;
      }
      if (op != 0)
      {
        switch (op)
        {
          case '<':
          case '>': cout << (char) op;
        }
        cout << (char) op << endl << b;
        if (md != 0)
        {
          cout  << endl << (char) md << endl << c;
        }
        cout << endl;
      }
      cout << '=' << endl;
    }

    total_w += SetTimer (timer);

    // Obtaining the result.

    if (op != 0)
    for (int n = 0; n < nRepeat; n++)
    {
      if(!n)cerr << "Computing ";

      x = a;           // Result 1
      y = 0;           // Result 2

      if (md == 0)
      switch (op)
      {
        case  +1: if(!n)cerr <<"++a";       ++x;                break;
        case  -1: if(!n)cerr <<"--a";       --x;                break;
        case '+': if(!n)cerr <<"a+b";       x += b;             break;
        case '-': if(!n)cerr <<"a-b";       x -= b;             break;
        case '*': if(!n)cerr <<"a*b";       x *= b;             break;
        case '/': if(!n)cerr <<"a/b,a%b";   DivRem (x,y,a,b);   break;
        case '%': if(!n)cerr <<"a%b";       x %= b;             break;
        case '\\':if(!n)cerr <<"a\\b (pow)";power (x,a,lb);     break;
        case '<': if(!n)cerr <<"a<<b";      x <<= lb;           break;
        case '>': if(!n)cerr <<"a>>b";      x >>= lb;           break;
        case '|': if(!n)cerr <<"a|b (or)";  x |= b;             break;
        case '&': if(!n)cerr <<"a&b (and)"; x &= b;             break;
        case '^': if(!n)cerr <<"a^b (xor)"; x ^= b;             break;
        case 'V':
        case 'v': if(!n)cerr <<"sqrt(b)";   SqrRoot (x, b);     break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': if(!n)cerr <<"a+(b*c)";   x = a; x += b * c;  break;
        case '-': if(!n)cerr <<"a-(b*c)";   x = a; x -= b * c;  break;
      }
      else if (md == '%')
      switch (op)
      {
        case '\\':if(!n)cerr <<"a\\b%c";    PowerMod (x,a,b,c); break;
      }
      if(!n)cerr << endl;

      total_c += SetTimer (timer);
    }

    // Writing the result.

    if (op != 0)
    {
      cout << x << endl;
      switch (op)
      {
//      case 'V':
//      case 'v':
        case '/': cout << y << endl;
                  break;
      }

      total_w += SetTimer (timer);
    }
    else    // Just writing the operand
    {
      cout << a << endl;
      total_w += SetTimer (timer);
    }

    // Comparing by alternative operation.

    if (nCheck != 0 && op != 0 && alt != 0)
    {
      p = 0;        // Number to compare with result
      q = 0;        // Number to compare with result

      cerr << "Computing ";

      if (md == 0)
      switch (op)
      {
        case +1:  cerr <<"a+1";             p = a + 1;          break;
        case -1:  cerr <<"a-1";             p = a - 1;          break;
        case '+': cerr <<"a+b (alt)";       p = a + b;          break;
        case '-': cerr <<"a-b (alt)";       p = a - b;          break;
        case '*': cerr <<"a*b";             p = a * b;          break;
        case '/': cerr <<"a/b,a%b";         p = a / b;
                                            q = a % b;          break;
        case '%': cerr <<"a%b";             p = a % b;          break;
        case '\\':cerr <<"a\\b (pow)";      p = power (a,lb);   break;
        case '<': if (b >= 0)
                       { cerr <<"a*2\\b";   p = a * power2_ZZ (lb);  }
                  else { cerr <<"a/2\\-b";  p = a / power2_ZZ (-lb); }
                                                                break;
        case '>': if (b >= 0)
                       { cerr <<"a/2\\b";   p = a / power2_ZZ (lb);  }
                  else { cerr <<"a*2\\-b";  p = a * power2_ZZ (-lb); }
                                                                break;
//      case '|': cerr <<"~(~a&~b)";        p = ~(~a & ~b);     break;
//      case '&': cerr <<"~(~a|~b)";        p = ~(~a | ~b);     break;
//      case '^': cerr <<"~a&b|a&~b";       p = (~a & b) | (a & ~b);
                                                                break;
        case 'V':
        case 'v': cerr <<"sqrt(b)";         p = SqrRoot (b);    break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': cerr <<"a+(b*c)";         p = a + b * c;      break;
        case '-': cerr <<"a-(b*c)";         p = a - b * c;      break;
      }
      else if (md == '%')
      switch (op)
      {
        case '\\':cerr <<"(a\\b)%c";        p = power(a,lb) % c;
                                                                break;
      }
      cerr << endl;

      total_c += SetTimer (timer);

      cout << p << endl;
      switch (op)
      {
        case '/':
        case 'V':
        case 'v':   cout << q << endl;
                    break;
      }

      total_w += SetTimer (timer);

      cerr << "Comparing... ";

      if (x == p && y == q)
      {
        cerr << "OK";
      }
      else
      {
        cout << "NO MATCH" << endl;
        cerr << "FAIL!\b";
        retcode = 1;
      }
      cerr << endl;
    }

    // Comparing by reverse operation.

    if (nCheck != 0 && op != 0 && rev != 0)
    {
      p = x;       // Number to compare with operand
      q = 0;       // Number to compare with 0
      cmp = a;     // Operand to compare

      cerr << "Computing ";

      if (md == 0)
      switch (op)
      {
        case  +1: cerr << "--++a";      --p;                break;
        case  -1: cerr << "++--a";      ++p;                break;
        case '+': cerr << "a+b-b";      p -= b;             break;
        case '-': cerr << "a-b+b";      p += b;             break;
        case '*': cerr << "a*b/b";      p /= b;             break;
        case '/': cerr << "a/b*b+a%b";  p *= b; p += y;     break;
        case '%': cerr << "a%b+a/b*b";  p += (a/b)*b;       break;
        case '\\':if ((lb & 1) == 0 && cmp < 0) cmp = - cmp;
                  switch (lb) {
          case 8: cerr << "sqrt(a\\8)"; p = SqrRoot (p);
          case 4: cerr << "sqrt(a\\4)"; p = SqrRoot (p);
          case 2: cerr << "sqrt(a\\2)"; p = SqrRoot (p);
                  }                                         break;
        case '<': cerr << "(a<<b)>>b";  p >>= lb;           break;
        case '^': cerr << "a^b^b";      p ^= b;             break;
        case 'V':
        case 'v': cerr << "sqrt(b)*sqrt(b)+(b-sqrt(b)*sqrt(b))";
                               cmp = b; p = sqr(p); p += y; break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': cerr << "(a+b*c-a)/c";p = p - a;
                                        p /= c;
                                        cmp = b;            break;
        case '-': cerr << "(a-a-b*c)/c";p = a - p;
                                        p /= c;
                                        cmp = b;            break;
      }
      cerr << endl;

      total_c += SetTimer (timer);

      cout << p << endl;
      if (md == 0)
      switch (op)
      {
        case '*':
        case '\\':  cout << q << endl;
      }
      else if (md == '*')
      switch (op)
      {
        case '+':
        case '-':   cout << q << endl;
      }

      total_w += SetTimer (timer);

      cerr << "Comparing... ";

      if (cmp == p && q == 0)
      {
        cerr << "OK";
      }
      else
      {
        cout << "NO MATCH" << endl;
        cerr << "FAIL!\b";
        retcode = 1;
      }
      cerr << endl;
    }

    cout << endl;
    if (!cout)
    {
       cerr << "Write error" << endl;
       return 1;
    }
  }

  cout << dec;

  if (*pszGenerate == 0)
  {
    long total = 0;
    cout << "Reading    " << setw (6) << total_r << " ms\n"; total += total_r;
    cout << "Computing  " << setw (6) << total_c << " ms";   total += total_c;
    if (nRepeat > 1)
      cout << " (" << ((total_c * 2 + nRepeat) / (nRepeat * 2)) << " ms "
           << nRepeat << " times)";
    cout << "\n";
    cout << "Writing    " << setw (6) << total_w << " ms\n"; total += total_w;
    cout << "Total      " << setw (6) << total   << " ms\n";
  }

  if (!cout)
  {
     cerr << "Write error" << endl;
     return 1;
  }

  return retcode;
}

static int GenerateRandom (ZZ& n, size_t radix, size_t maxexp)
{
  return GenerateRandom (trunc_long(n,CHAR_BIT * sizeof (long)), radix, maxexp);
}

static int GenerateRandom (long n, size_t radix, size_t maxexp)
{
  assert (radix > 0);

  if (n < 0) { cout << '-'; n = -n; }

  cerr << "Generating random number of length " << n;
  if (maxexp != 0)
  {
    maxexp = (int) (dRand3() * maxexp);
    cerr << " and exponent " << maxexp;
  }
  cerr << endl;

  if (n != 0)
  {
    cout << (int) (1 + (dRand3() * (radix - 1)));
    while (--n != 0)
    {
      cout << (int) (dRand3() * radix);
    }
    for (; maxexp > 0; --maxexp)
    {
      cout << 0;
    }
  }
  else
  {
    cout << 0;
  }
  return 0;
}

static void ShowStat (ZZ& a, ZZ& na)
{
/*cerr << (const char*) cHexDump (a.hiword());
  if (a.length() > 1)
  {
    cHexDump a2 (a.code() [a.length() - 2]);
    ((char*) (const char*) a2) [2] = 0;
    cerr <<" "<< (const char*) a2 <<"..";
  }
  else
    cerr <<"     ";

  cerr <<": "<< setw (7) << na <<" bit" << (na != 1?"s":" ");
  size_t la = a.words();
  cerr <<" in "<< setw (5) << (long)la <<" word" << ("s" + (la == 1));
  cerr << endl;
*/}
